/*
 * WScout - Lightweight PCAP visualizer.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <string>
#include <sstream>
#include <iomanip>
#include <locale>
#include <stdexcept>
#include <cassert>

#include "wscout_hexdump.hh"

namespace wscout
{

  namespace
  {

    static
    std::ostream&
    dump(std::ostream& os, char c)
    {
      if (std::isprint(c))
	switch (c)
	  {
	  case '<':
	    return os << "&lt;";
	  case '>':
	    return os << "&gt;";
	  default:
	    return os << c;
	  }
      else
	return os << '.';
    }

  } // End of anonymous namespace.

  std::string
  hexdump(const void* data, size_t len)
  {
    const unsigned		lwidth = 16;	// Line width.
    const unsigned		bwidth = 4;	// Block width.
    const std::string		bigsep = "   ";	// Big separator.
    const std::string		smallsep = " ";	// Small separator.
    const std::string		colsep = " <b>|</b> ";	// Column separator.

    const unsigned char* const	b = static_cast<const unsigned char*> (data);
    const unsigned char* const	e = b + len;

    std::ostringstream		os;

    os << std::hex << std::uppercase << "<pre>";
    for (const unsigned char* c = b; c < e; c += lwidth)
      {
	os << "<b>" << std::setfill ('0') << std::setw (8) << (c - b) << "</b>";
	for (unsigned i = 0; i < lwidth; i += bwidth)
	  {
	    os << bigsep;
	    for (unsigned j = 0; j < bwidth and c + i + j < e; ++j)
	      {
		if (j)
		  os << smallsep;
		os << std::setfill ('0') << std::setw (2)
		   << unsigned (c[i + j]);
	      }
	    for (int j = bwidth - 1; j >= 0 and c + i + j >= e; --j)
	      {
		if (j)
		  os << smallsep;
		os << "  ";
	      }
	  }
	os << bigsep << colsep;
	for (unsigned i = 0; i < lwidth; ++i)
	  dump(os, c + i < e ? c[i] : ' ');
	os << colsep << '\n';
      }
    os << "</pre>";
    return os.str();
  }

  std::vector<uint8_t>
  unhexdump(const std::string& s)
  {
    typedef std::vector<uint8_t> vector;

    if (s.size() % 2)
      throw std::invalid_argument ("Hexadecimal strings must have an even "
				   "number of digits");

    const size_t		size = s.size() / 2;

    vector			res (size);
    std::string::const_iterator	cur = s.begin();

    for (size_t i = 0; i < size; ++i)
      {
	uint8_t	t = 0;

	for (size_t j = 0; j < 2; ++j, ++cur)
	  {
	    assert(cur != s.end());
	    t = t << 4;
	    if ('0' <= *cur and *cur <= '9')
	      t |= *cur - '0';
	    else if ('a' <= *cur and *cur <= 'f')
	      t |= *cur - 'a' + 10;
	    else if ('A' <= *cur and *cur <= 'F')
	      t |= *cur - 'A' + 10;
	    else
	      throw std::invalid_argument ("Not an hexadecimal string");
	  }
	res[i] = t;
      }
    assert(cur == s.end());
    return res;
  }


} // End of namespace wscout.


