/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_OPTIONS_HXX_
# define TOOL_OPTIONS_HXX_

# include <boost/algorithm/string/classification.hpp>
# include <boost/algorithm/string/split.hpp>
# include <boost/lexical_cast.hpp>
# include <boost/preprocessor/seq/for_each.hpp>
# include <boost/preprocessor/seq/for_each_i.hpp>
# include <boost/preprocessor/tuple/elem.hpp>
# include <boost/preprocessor/stringize.hpp>

# include <cassert>
# include <stdexcept>
# include <sstream>
# include <unistd.h>

# include "options.hh"

#ifdef HAVE_OPENSSL
# include <wipal/wifi/frame/unique_id/hash.hh>
#endif // HAVE_OPENSSL
#include <wipal/wifi/frame/unique_id/timestamp.hh>
#include <wipal/wifi/frame/unique_id/addr_timestamp.hh>
#include <wipal/wifi/frame/unique_id/source_bssid_timestamp.hh>
#include <wipal/wifi/frame/unique_id/seqctl_timestamp.hh>
#include <wipal/wifi/frame/unique_id/seqctl_bssid_timestamp.hh>
#include <wipal/wifi/frame/unique_id/seqctl_dest_bssid_timestamp.hh>
#include <wipal/wifi/frame/unique_id/seqctl_source_bssid_timestamp.hh>

namespace tool
{

  namespace opt
  {

    namespace internals
    {

      inline
      void
      usage(const std::string& progname,
	    const std::string& available_options,
	    unsigned ninputs, unsigned noutputs)
      {
	std::cerr << "Invocation:\n\t" << progname << " [options]";
	switch (ninputs)
	  {
	  case 0:					break;
	  case 1:  std::cerr << " <input>";		break;
	  case 2:  std::cerr << " <input-1> <input-2>";	break;
	  default: std::cerr << " <inputs...>";		break;
	  }
	switch (noutputs)
	  {
	  case 0:					break;
	  case 1:  std::cerr << " <output>";		break;
	  default: std::cerr << " <outputs...>";	break;
	  }
	std::cerr << "\n\nOptions:\n";

# define WP_help_print_opt(R, Unused, Opt)				\
	if (available_options.find(BOOST_PP_CAT(WP_OPT_, Opt)) <	\
	    available_options.size())					\
	  std::cerr << BOOST_PP_CAT(WP_HELP_, Opt);

	BOOST_PP_SEQ_FOR_EACH(WP_help_print_opt, ~, WP_OPT_LIST)

# undef WP_help_print

# define WP_help_attr_string(R, Unused, A)				\
	  "  " BOOST_PP_STRINGIZE(A) ":\t" BOOST_PP_CAT(WP_ATTR_HELP_, A) "\n"
# define WP_help_linktypes_string(R, Unused, LT)			\
	"  " BOOST_PP_STRINGIZE(BOOST_PP_TUPLE_ELEM(2, 0, LT)) "\n"

	if (available_options.find(WP_OPT_ATTR) < available_options.size())
	  std::cerr << "\nAvailable attributes:\n"
	    BOOST_PP_SEQ_FOR_EACH(WP_help_attr_string, ~, WP_ATTRIBUTES);
	std::cerr << "\nSupported PCAP link types:\n"
	  BOOST_PP_SEQ_FOR_EACH(WP_help_linktypes_string, ~, WP_LINKTYPES);

# undef WP_help_linktypes_string
# undef WP_help_attr_string
      }

      inline
      void
      version(const std::string& progname)
      {
	std::cerr << progname << "\n"
	  "Trace Tools, version " << PACKAGE_VERSION << '.' << std::endl;
      }

      inline
      std::pair<unsigned, unsigned>
      make_reference(const std::string& s)
      {
	static const std::string err = ("Invalid reference frame format. "
					"Format is ``<unsigned>-<unsigned>''");

	std::vector<std::string>	v;

	boost::split(v, s, boost::is_any_of("-"));
	if (v.size() != 2)
	  throw std::invalid_argument (err);
	try
	  {
	    return std::make_pair(boost::lexical_cast<unsigned> (v[0]),
				  boost::lexical_cast<unsigned> (v[1]));
	  }
	catch (const boost::bad_lexical_cast&)
	  {
	    throw std::invalid_argument (err);
	  }
      }

    }

    inline
    options
    parse_args(int argc, char* const* argv,
	       const std::string& available_options,
	       int ninputs, int noutputs)
    {
      const std::string optstring =
	available_options + WP_OPT_HELP WP_OPT_VERSION_;

      options	r;
      int	o;

      r.mapping = WP_MAPPING_DEFAULT;
      r.attributes = BOOST_PP_STRINGIZE(WP_DEFAULT_ATTRIBUTE);
      r.endianness = endian::guess;

      r.cmp = options::cmp_default;
      r.prism_noise_means_error = false;
      r.print_errors = false;
      r.print_frame_indices = true;
      r.print_column_headers = isatty(STDOUT_FILENO);
      r.print_microseconds = false;
      r.print_summary = true;

      while ((o = getopt(argc, argv, optstring.c_str())) != -1)
	switch (o)
	  {
	  case '8':
	    r.cmp = options::cmp_802;
	    break;

	  case 'a':
	    r.attributes = optarg;
	    break;

	  case 'b':
	    r.cmp = options::cmp_bytes;
	    break;

	  case 'c':
	    r.print_column_headers = false;
	    break;

	  case 'C':
	    r.print_column_headers = true;
	    break;

	  case 'd':
	    r.cmp = options::cmp_default;
	    break;

	  case 'e':
	    r.print_errors = false;
	    break;

	  case 'E':
	    r.print_errors = true;
	    break;

	  case 'h':
	    internals::usage(argv[0], optstring, ninputs, noutputs);
	    exit(EXIT_SUCCESS);
	    break;

	  case 'i':
	    r.print_frame_indices = false;
	    break;

	  case 'I':
	    r.print_frame_indices = true;
	    break;

	  case 'm':
	    r.mapping = optarg;
	    break;

	  case 'n':
	    r.endianness = endian::little;
	    break;

	  case 'N':
	    r.endianness = endian::big;
	    break;

	  case 'p':
	    r.prism_noise_means_error = false;
	    break;

	  case 'P':
	    r.prism_noise_means_error = true;
	    break;

	  case 'q':
	    r.print_errors = false;
	    r.print_frame_indices = false;
	    r.print_column_headers = false;
	    r.print_microseconds = false;
	    r.print_summary = false;
	    break;

	  case 'r':
	    try
	      {
		r.ref_blacklist.insert(internals::make_reference(optarg));
	      }
	    catch (const std::invalid_argument& i)
	      {
		std::cerr << "ERROR: " << i.what() << '.' << std::endl;
		exit(EXIT_FAILURE);
	      }
	    break;

	  case 't':
	    r.cmp = options::cmp_time;
	    break;

	  case 'u':
	    r.print_microseconds = false;
	    break;

	  case 'U':
	      r.print_microseconds = true;
	      break;

	  case 'v':
	    internals::version(argv[0]);
	    exit(EXIT_SUCCESS);

	  default: /* '?' */
	    internals::usage(argv[0], optstring, ninputs, noutputs);
	    exit(EXIT_FAILURE);
	  }

      assert(ninputs  != options::arbitrary_number_of_inputs or
	     noutputs != options::arbitrary_number_of_outputs);

      if (ninputs == options::arbitrary_number_of_inputs)
	ninputs = argc - optind - noutputs;
      if (noutputs == options::arbitrary_number_of_outputs)
	noutputs = argc - optind - ninputs;
      if (ninputs < 0 or noutputs < 0 or optind + ninputs + noutputs != argc)
	{
	  std::cerr << "ERROR: Wrong number of arguments." << std::endl;
	  internals::usage(argv[0], optstring, ninputs, noutputs);
	  exit(EXIT_FAILURE);
	}
      for (int i = 0; i < ninputs; ++i)
	r.input.push_back(argv[optind + i]);
      for (int i = 0; i < noutputs; ++i)
	r.output.push_back(argv[optind + ninputs + i]);

      return r;
    }

    template <class Fun>
    int
    dispatch_unique_id(const options& o, Fun f)
    {
      if (false)
	;
# define WP_dispatch_unique_id_else(R, Unused, A)			\
      else if (o.attributes == BOOST_PP_STRINGIZE(A))			\
	return f.template operator () <wifi::frame::BOOST_PP_CAT(A, _id)>(o);

      BOOST_PP_SEQ_FOR_EACH(WP_dispatch_unique_id_else, ~, WP_ATTRIBUTES)

# undef WP_dispatch_unique_id_else

      throw std::invalid_argument ("Invalid distinctive attributes");
    }

    inline
    void
    unsupported_linktype(const pcapxx::descriptor<>::link_type lt)
    {
      std::ostringstream	os;

# define WP_linktype_string(R, Unused, I, LT)		\
      BOOST_PP_IF(I, ", ", "")				\
      BOOST_PP_STRINGIZE(BOOST_PP_TUPLE_ELEM(2, 0, LT))

      os << "Unsupported PCAP link type: " << lt << " - expected one of ["
	BOOST_PP_SEQ_FOR_EACH_I(WP_linktype_string, ~, WP_LINKTYPES) "]";

# undef WP_linktype_string

      throw std::invalid_argument (os.str());
    }

  } // End of namespace tool::opt.

} // End of namespace tool.

#endif // TOOL_OPTIONS_HXX_
