/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_STATS_STATS_HXX_
# define WIFI_FRAME_STATS_STATS_HXX_

# include <wipal/wifi/frame/stats/stats.hh>
# include <wipal/wifi/frame/dissector/dissector.hh>

namespace wifi
{

  namespace frame
  {

    namespace stats
    {

      /*-------------.
      | Constructors |
      `-------------*/

      inline
      stats::stats(): me_ (sc_)
      {
      }

      inline
      stats::hooks::hooks(stats& parent): type (-1),
					  subtype (-1),
					  retx (false),
					  txer_type (unknown),
					  receiver (0),
					  parent (&parent)
      {
      }

      inline
      stats::state::state(unsigned			last_frame_id,
			  const seqctl&			last_sc,
			  const tool::microseconds&	last_rx_stamp):
	last_frame_id (last_frame_id),
	last_sc (last_sc),
	last_rx_stamp (last_rx_stamp)
      {
      }

      /*-------.
      | Misc.  |
      `-------*/

      inline
      void
      stats::restart()
      {
	states_.clear();
	df_.restart();
	tp_.restart();
	ay_.restart();
	ag_.restart();
	bg_.restart();
	ig_.restart();
	sg_.restart();
	pg_.restart();
      }

      inline
      std::ostream&
      stats::print(std::ostream& o) const
      {
	return
	  o << df_ << me_ << sc_ << '\n'
	    << "begin gap length frequencies\n" << gl_
	    << "end gap length frequencies\n\n"
	    << "begin T-Fi plot\n" << tp_ << "end T-Fi plot\n\n"
	    << "begin BSS figures\n" << bs_ << "end BSS figures\n\n"
	    << "begin SSID figures\n" << ss_ << "end SSID figures\n\n"
	    << "begin activity\n" << ay_ << "end activity\n\n"
	    << "begin MAC addr. growth\n" << ag_ << "end MAC addr. growth\n\n"
	    << "begin BSSID growth\n" << bg_ << "end BSSID growth\n\n"
	    << "begin IBSSID growth\n" << ig_ << "end IBSSID growth\n\n"
	    << "begin SSID growth\n" << sg_ << "end SSID growth\n\n"
	    << "begin AP growth\n" << pg_ << "end AP growth\n"
	    << std::flush;
      }

      /*-----------------.
      | Stats. accessors |
      `-----------------*/

      inline
      const delimiter_frames&
      stats::delimiter_frames_get() const
      {
	return df_;
      }

      inline
      const simple_counters&
      stats::simple_counters_get() const
      {
	return sc_;
      }

      inline
      const gap_lengths&
      stats::gap_lengths_get() const
      {
	return gl_;
      }

      inline
      const missed_estimations&
      stats::missed_estimations_get() const
      {
	return me_;
      }

      inline
      const tfi_plot&
      stats::tfi_plot_get() const
      {
	return tp_;
      }

      inline
      const bss_stats&
      stats::bss_stats_get() const
      {
	return bs_;
      }

      inline
      const ssid_stats&
      stats::ssid_stats_get() const
      {
	return ss_;
      }

      inline
      const activity&
      stats::activity_get() const
      {
	return ay_;
      }

      inline
      const stats::addr_growth&
      stats::addr_growth_get() const
      {
	return ag_;
      }

      inline
      const stats::bssid_growth&
      stats::bssid_growth_get() const
      {
	return bg_;
      }

      inline
      const stats::ssid_growth&
      stats::ssid_growth_get() const
      {
	return sg_;
      }

      inline
      const stats::ap_growth&
      stats::ap_growth_get() const
      {
	return pg_;
      }

      /*-----------------------------.
      | Frame parsing and accounting |
      `-----------------------------*/

      inline
      void
      stats::hooks::invalid_type_or_subtype_hook(const void* f, size_t,
						 dissector_status::status)
      {
	type = type_of(f);
	subtype = subtype_of(f);
	retx = control_flag(3, f);
      }

      inline
      void
      stats::hooks::addr_hook(const void* f, size_t,
			      unsigned i, const addr& a)
      {
	bool	account_addr = true;

	switch (i)
	  {
	  case 1:
	    type = type_of(f);
	    subtype = subtype_of(f);
	    retx = control_flag(3, f);
	    receiver = &a;

	    switch (type) // Do not account BSSIDs as MAC addresses.
	      {
	      case type::control:
		account_addr = subtype != ctl::subtype::ps_poll;	break;
	      case type::data:
		account_addr = not to_ds(f) or from_ds(f);		break;
	      default:							break;
	      }
	    break;

	  case 2:
	    txer_type =
	      parent->aps_.find(a) == parent->aps_.end() ? client : ap;

	    if (type == frame::type::management or type == frame::type::data)
	      id = state_id (a, addr::broadcast());

	    switch (type) // Do not account BSSIDs as MAC addresses.
	      {
	      case type::control:
		account_addr =
		  subtype != ctl::subtype::cf_end and
		  subtype != ctl::subtype::cf_end_cf_ack;	break;
	      case type::data:
		account_addr = to_ds(f) or not from_ds(f);	break;
	      default:						break;
	      }
	    break;

	  case 3:
	    // Do not account BSSIDs as MAC addresses.
	    switch (type)
	      {
	      case type::data:   account_addr = to_ds(f) or from_ds(f);	break;
	      case type::management:    account_addr = false;		break;
	      default:							break;
	      }
	    break;
	  }

	if (account_addr)
	  parent->account_addr(a);
      }

      inline
      void
      stats::hooks::seq_ctl_hook(const void*, size_t, unsigned fn, unsigned sn)
      {
	sc = seqctl (sn, fn);
      }

      inline
      void
      stats::hooks::qos_ctl_hook(const void*, size_t,
				 unsigned tid,
				 bool, unsigned, unsigned)
      {
	assert(id);

	if (*receiver == addr::broadcast())
	  return;

	id->get<1>() = *receiver;
	id->get<2>() = tid;
      }

      inline
      void
      stats::hooks::beacon_hook(const mgt::header* header, size_t caplen)
      {
	beacon = beacon_header (header, caplen);

	const beacon_body*  bb =
	  reinterpret_cast<const beacon_body*> (header + 1);

	if (caplen < sizeof (*header) + sizeof (*bb))
	  return;

	if ((tool::extract_little_endian_short_u(bb->capability) & 0x3) == 1)
	  {
	    txer_type = ap;
	    parent->aps_.insert(header->addrs[mgt::header::sa]);
	    parent->aps_.insert(header->addrs[mgt::header::bssid]);
	  }
      }

      template <class Phy, class Frame>
      void
      stats::account(const Frame& f)
      {
	const stats_dissector	d = dissect<Phy, hooks>(f, hooks (*this));

	if (d.status() == dissector_status::invalid)
	  return;

	account_frame<Phy>(f, d);
	if (d.beacon)
	  account_beacon(f.microseconds(), d.beacon->first, d.beacon->second,
			 d.txer_type == client);

	if (d.id)
	  update_states(f, d);
      }

      template <class Frame>
      void
      stats::update_states(const Frame& f, const stats_dissector& d)
      {
	assert(d.id);
	assert(d.sc);

	const unsigned			frame_id = f.id();
	const tool::microseconds&	rx_stamp = f.microseconds();
	const state_map::iterator	i = states_.find(*d.id);

	state				new_state (frame_id, *d.sc, rx_stamp);

	// First occurence of this state ID.
	if (i == states_.end())
	  {
	    states_.insert(std::make_pair(*d.id, new_state));
	    return;
	  }

	const state&			s = i->second;
	const tool::microseconds	delta_t = rx_stamp - s.last_rx_stamp;

	if (state_lifetime < delta_t)
	  account_expired();
	else if (*d.sc < s.last_sc)
	  account_out_of_order();
	else if (*d.sc == s.last_sc)
	  {
	    if (not d.retx)
	      account_duplicate();
	  }
	else
	  {
	    assert(*d.sc > s.last_sc);

	    if (d.retx) // We missed the original non-retransmitted frame.
	      account_miss();

	    const int	delta_s = d.sc->seqnum_get() - s.last_sc.seqnum_get();

	    /* If there is a sequence number gap, *and* its value is
	     * credible, account for it.
	     *
	     * Credible means the STA could have sent all the missed
	     * frames since we last seen it.  E.g. if it takes 100us
	     * to send the smallest allowed frame, and the last frame
	     * we saw is 300us old, then only gaps of length 1 will be
	     * credible.
	     */
	    if (1 < delta_s)
	      {
		if (delta_s < delta_t / min_frame_sending_time)
		  account_gap(delta_s - 1);
		else
		  account_seq_anomaly();
	      }
	  }
	i->second = new_state;
      }

      template <class PhyHeader, class Frame>
      void
      stats::account_frame(const Frame&	f, const stats_dissector& d)
      {
	const tool::microseconds tstamp = f.microseconds();
	const bool		 from_ap = d.txer_type == ap;
	const pcapxx::pkthdr&	 pcap = *f.pcap_header();
	const PhyHeader*	 phy = reinterpret_cast<const PhyHeader*>
				       (f.bytes().get());
	const size_t		 phy_len = phy->len(pcap.caplen, f.swapped());

	size_t			 len;

	if (phy_len > pcap.len)
	  {
	    std::cerr << "WARNING: Packet " << f
		      << " too short to make sense.\n"
		      << "WARNING: Packet length is " << pcap.len << ".\n"
		      << "WARNING: PHY header length should be " << phy_len
		      << '.' << std::endl;
	    len = 0;
	  }
	else
	  len = pcap.len - phy_len;

	df_.account_frame(tstamp, pcap.ts);
	sc_.account_frame(len, d.type, d.subtype, d.retx, from_ap);
	ay_.account_frame(tstamp, len, d.type, d.subtype, from_ap);
	tp_.account_frame(f.id(), tstamp);

	// Update growth stats.
	ag_.update(tstamp);	// MAC addr.
	bg_.update(tstamp);	// BSSIDs.
	ig_.update(tstamp);	// IBSSIDs.
	sg_.update(tstamp);	// SSIDs.
	pg_.update(tstamp);	// APs.
      }

      inline
      void
      stats::account_beacon(const tool::microseconds&	tstamp,
			    const mgt::header*		header,
			    size_t			caplen,
			    bool			ibss)
      {
	if (caplen < sizeof (*header))
	  return;

	const wifi::addr&	ap = header->addrs[mgt::header::sa];
	const wifi::addr&	bss = header->addrs[mgt::header::bssid];

	me_.account_beacon(tstamp, ap);
	bs_.account_beacon(bss, ap);
	if (ibss)
	  ig_.update(bss);
	else
	  bg_.update(bss);
	pg_.update(ap);

	const beacon_body*  bb =
	  reinterpret_cast<const beacon_body*> (header + 1);

	if (caplen < sizeof (*header) + sizeof (*bb) or
	    caplen < sizeof (*header) + sizeof (*bb) + bb->ssid_len)
	  return;

	const std::string	ssid (bb->ssid, bb->ssid + bb->ssid_len);

	ss_.account_beacon(ssid);
	sg_.update(ssid);
      }

      inline
      void
      stats::account_miss()
      {
	me_.account_miss();
	tp_.account_miss();
      }

      inline
      void
      stats::account_gap(unsigned gaplen)
      {
	sc_.account_gap();
	gl_.account_gap(gaplen);
	me_.account_gap(gaplen);
	tp_.account_gap(gaplen);
      }

      inline
      void
      stats::account_expired()
      {
	sc_.account_expired();
      }

      inline
      void
      stats::account_out_of_order()
      {
	sc_.account_out_of_order();
      }

      inline
      void
      stats::account_duplicate()
      {
	sc_.account_duplicate();
      }

      inline
      void
      stats::account_seq_anomaly()
      {
	sc_.account_seq_anomaly();
      }

      inline
      void
      stats::account_addr(const addr& a)
      {
	ag_.update(a);
      }

    } // End of namespace wifi::frame::stats.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_STATS_STATS_HXX_
