/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_MACTIME_TRACKER_HH_
# define WIFI_MACTIME_TRACKER_HH_

# include <wipal/pcap/frame_descriptor.hh>
# include <wipal/wifi/frame/frame.hh>

extern "C"
{
# include <inttypes.h>
}

# include <wipal/tool/static_base.hh>
# include <wipal/tool/microseconds.hh>
# include <wipal/phy/prism_header.hh>

namespace wifi
{

  /**
   * Microsecond-grained time tracker.
   *
   * PHY headers have various fields that holds frames' arrival time
   * on the medium.  Some of them are only 32 bits long, thus wrapping
   * every one hour and a half (roughly).  This class tracks the
   * evolution of such fields and maintains a microsecond count that
   * does not wrap.
   *
   * Just call the tick() method for every frames you get in your
   * trace.
   */
  template <class Bottom = tool::bottom>
  struct mactime_tracker: WP_INHERIT(public tool::static_base,
				     mactime_tracker<Bottom>)
  {
    /// Default constructor - initialize the microsecond count to zero.
    mactime_tracker();

    /**
     * Update the tracker with the given unsigned integer timestamp.
     *
     * This method tracks wraparounds and maintains a consistent
     * microsecond count.
     */
    template <class UIntType>
    void
    tick_uint(UIntType ts);

    /**
     * Update the tracker with the given @c timeval structure.
     *
     * This method assumes @c timeval structures never wrap around.
     * A negative structure (@c tv_sec < 0) will yield valid but
     * negative microsecond counts).
     */
    void			tick_tv(const struct timeval& tv);

    /// Update the tracker according to the given frame.
    template <class HeaderType, class D>
    void
    tick(const pcapxx::frame_descriptor<D>&	frame,
	 tool::endian::endianness		phy_end);

    /// Retrieve the current microsecond count without wraps.
    tool::microseconds		microseconds() const;

  private:
    uint64_t			last_stamp_;
    tool::microseconds		ms_count_;
  };

} // End of namespace wifi.

# include "mactime_tracker.hxx"

#endif // ! WIFI_MACTIME_TRACKER_HH_
