/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_TIME_CONVERTER_HH_
# define WIFI_TIME_CONVERTER_HH_

extern "C"
{
# include <inttypes.h>
}

# include <wipal/tool/static_base.hh>
# include <wipal/tool/microseconds.hh>

namespace wifi
{

  /**
   * Convert a @c time_tracker microsecond count to various PCAP and
   * PHY fields.
   *
   * Use the @c time_tracker class to have a global timer that tracks
   * frame arrivals.  Use this class to convert @c time_tracker values
   * back into various time fields (e.g. PCAP time, Prism @c mactime,
   * etc.).
   *
   * In order to work properly, this class must be able to map
   * a reference microsecond count to reference time fields (PCAP
   * time, host time, MAC time).  Thus its constructor.
   *
   * @see time_tracker.
   */
  template <class PhyHeaderType, class Bottom = tool::bottom>
  struct time_converter: WP_INHERIT(public tool::static_base,
				    time_converter<PhyHeaderType, Bottom>)
  {
    /// @name Static hierarchies related types
    //@{
    typedef WP_GET_EXACT(Bottom, time_converter<PhyHeaderType, Bottom>)
								exact_type;
    typedef tool::static_base<exact_type>			super_type;
    //@}

    /// Type of PCAP headers' timing information.
    typedef struct timeval			pcap_time_type;

    /// Type of PHY headers' timing information.
    typedef typename PhyHeaderType::time_type	phy_time_type;

    /// Construct a converter for some given reference times.
    time_converter(const pcap_time_type&	local_pcap_time,
		   const phy_time_type&		local_phy_time,
		   const tool::microseconds&	global_time);

    /// Hold @c time_converter 's conversion results.
    struct time_values
    {
      /// Construct an non-initialized @c time_values object.
      time_values();

      /// Construct an initialized @c time_values object.
      time_values(const pcap_time_type&	pcap_time,
		  const phy_time_type&	phy_time);

      /// PCAP headers' time.
      pcap_time_type	pcap_time;

      /// PHY headers' time.
      phy_time_type	phy_time;
    };

    /// Convert from a global microsecond count back to time values.
    time_values	operator () (const tool::microseconds& global_time) const;

  private:
    time_values		local_reference_;
    tool::microseconds	global_reference_;
  };

} // End of namespace wifi.

# include "time_converter.hxx"

#endif // ! WIFI_TIME_CONVERTER_HH_
