/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <iostream>
#include <stdexcept>
#include <cstdlib>

#include <boost/format.hpp>
#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include <gmpxx.h>

#include "config.hh"

#include <wipal/tool/options.hh>
#include <wipal/wifi/frame/filter/linear_regression_synchronizer.hh>
#include <wipal/wifi/frame/filter/intersector.hh>
#include <wipal/wifi/frame/filter/fast_intersector.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/addr_mapping.hh>

namespace
{

  struct sync_unique_frames_dumper
  {

    sync_unique_frames_dumper(const tool::opt::options& options):
      options_ (options),
      coefs_ ("%10s %15s"),
      indices_ (" %10s %10s"),
      msecs_   (" %15s %15s"),
      errors_  (" %10s")
    {
    }

    void	print_colheads()
    {
      std::cout << (coefs_ % "a" % "b");
      if (options_.print_frame_indices)
	std::cout << (indices_ % "frame 1" % "frame 2");
      if (options_.print_microseconds)
	std::cout << (msecs_ % "time 1 (us)" % "time 2 (us)");
      if (options_.print_errors)
	std::cout << (errors_ % "error");

      std::cout << '\n'
		<< (coefs_ % "=" % "=");
      if (options_.print_frame_indices)
	std::cout << (indices_ % "=======" % "=======");
      if (options_.print_microseconds)
	std::cout << (msecs_ % "===========" % "===========");
      if (options_.print_errors)
	std::cout << (errors_ % "=====");
      std::cout << std::endl;
    }

    template <class LRSync>
    void	operator () (const LRSync& s)
    {
      typedef typename LRSync::const_iterator	iterator;

      if (options_.print_column_headers)
	print_colheads();

      const iterator	end = s.end();

      for (iterator j = s.begin(); j != end; ++j)
	{
	  std::cout << (coefs_ % j->coefs.first % j->coefs.second);
	  if (options_.print_frame_indices)
	    std::cout << (indices_ % j->first % j->second);
	  if (options_.print_microseconds)
	    std::cout << (msecs_ % j->first.microseconds()
			         % j->second.microseconds());
	  if (options_.print_errors)
	    std::cout << (errors_ % (j->first.microseconds() * j->coefs.first +
				     j->coefs.second -
				     j->second.microseconds()));
	  std::cout << '\n';
	}
    }

  private:
    const tool::opt::options&	options_;
    boost::format		coefs_;
    boost::format		indices_;
    boost::format		msecs_;
    boost::format		errors_;
  };

  struct main_helper
  {
    template <class UniqueId>
    int
    operator () (const tool::opt::options& options) const
    {
      using wifi::frame::filter::provide_lr_synchronizer;

      mpf_set_default_prec(WP_GMP_FLOAT_PRECISION);

      wifi::addr_mapping	mapping (options.mapping);
      pcapxx::list_of_traces<>	traces1 (options.input[0]);
      pcapxx::list_of_traces<>	traces2 (options.input[1]);
      sync_unique_frames_dumper	dump (options);

      if (options.prism_noise_means_error)
	{
	  traces1.expect(pcapxx::descriptor<>::PRISM_HEADER);
	  traces2.expect(pcapxx::descriptor<>::PRISM_HEADER);
	}

      if (traces1.empty() or traces2.empty())
	return EXIT_SUCCESS;

      switch (traces1.get(0).linktype())
	{

#define WP_case(R, Unused, LT)						\
	case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	  traces1.expect(pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT)); \
	  traces2.expect(pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT)); \
	  provide_lr_synchronizer<UniqueId, BOOST_PP_TUPLE_ELEM(2, 1, LT), \
	    wifi::frame::filter::fast_intersector>			\
	    (traces1.begin(), traces1.end(),				\
	     traces2.begin(), traces2.end(),				\
	     mapping, dump,						\
	     options.prism_noise_means_error,				\
	     options.endianness,					\
	     options.ref_blacklist);					\
	  break;

	  BOOST_PP_SEQ_FOR_EACH(WP_case, ~, WP_LINKTYPES)

#undef WP_case

	default:
	  tool::opt::unsupported_linktype(traces1.get(0).linktype());
	}
      return EXIT_SUCCESS;
    }

  }; // End of struct main_helper

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  try
    {
      const tool::opt::options	options =
	tool::opt::parse_args(argc, argv,
			      WP_OPT_ATTR
			      WP_OPT_COLHEADS
			      WP_OPT_ERRS
			      WP_OPT_FRM_IDX
			      WP_OPT_MAPPING
			      WP_OPT_ENDIANNESS
			      WP_OPT_PRISM_NOISE
			      WP_OPT_USEC,
			      2, 0);
      return tool::opt::dispatch_unique_id(options, main_helper ());
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
